@@ -32,17 +32,19 @@ using L<JSON::Unmarshal|https://github.com/tadzik/JSON-Unmarshal>.
 
 It only outputs the "public" attributes (that is those with accessors
 created by declaring them with the '.' twigil. Attributes without acccessors
-are ignored.  
+are ignored.
 
 If you want to ignore any attributes without a value you can use the
-:skip-null adverb to C<marshal>, which will supress the marshalling of
-any undefined attributes.  Additionally if you want a finer-grained control
-over this behaviour there is a 'json-skip-null' attribute trait which
-will cause the specific attribute to be skipped if it isn't defined irrespective
-of the C<skip-null>.
+:skip-null adverb to C<marshal>, which will supress the marshalling
+of any undefined attributes.  Additionally if you want a finer-grained
+control over this behaviour there is a 'json-skip-null' attribute trait
+which will cause the specific attribute to be skipped if it isn't defined
+irrespective of the C<skip-null>.  C<skip-null> or the C<json-skip-null>
+trait is applied to a C<Positional> or C<Associative> attribute this
+will suppress the marshalling of an empty list or object attribute.
 
 To allow a finer degree of control of how an attribute is marshalled an
-attribute trait C<is marshalled-by> is provided, this can take either 
+attribute trait C<is marshalled-by> is provided, this can take either
 a Code object (an anonymous subroutine,) which should take as an argument
 the value to be marshalled and should return a value that can be completely
 represented as JSON, that is to say a string, number or boolean or a Hash
@@ -55,9 +57,9 @@ above.
 
 use JSON::Name;
 
-module JSON::Marshal:ver<0.0.6>:auth<github:jonathanstowe> {
+module JSON::Marshal:ver<0.0.16>:auth<github:jonathanstowe>:api<1.0> {
 
-    use JSON::Tiny;
+    use JSON::Fast:ver(v0.4+);
 
 
     role CustomMarshaller {
@@ -80,7 +82,7 @@ module JSON::Marshal:ver<0.0.6>:auth<github:jonathanstowe> {
         has Str $.marshaller is rw;
         method marshal($value, Mu:D $type) {
             my $meth = self.marshaller;
-            $value."$meth"();
+            $value.defined ?? $value."$meth"() !! $type;
         }
     }
 
@@ -100,33 +102,57 @@ module JSON::Marshal:ver<0.0.6>:auth<github:jonathanstowe> {
     multi sub trait_mod:<is> (Attribute $attr, :$json-skip-null!) is export {
         $attr does SkipNull;
     }
-    
-    multi sub _marshal(Cool $value) {
+
+    # This is a hard skip the attribute will never be emitted.
+    role JsonSkip {
+    }
+
+    multi sub trait_mod:<is> (Attribute $attr, :$json-skip!) is export {
+        $attr does JsonSkip;
+    }
+
+
+
+    multi sub _marshal(Cool $value, Bool :$skip-null) {
         $value;
     }
 
-    multi sub _marshal(%obj) returns Hash {
+    multi sub _marshal(Associative:U $, Bool :$skip-null --> Nil ) {
+        Nil;
+    }
+    multi sub _marshal(%obj, Bool :$skip-null --> Hash ) {
         my %ret;
 
         for %obj.kv -> $key, $value {
-            %ret{$key} = _marshal($value);
+            %ret{$key} = _marshal($value, :$skip-null);
         }
 
         %ret;
     }
 
-    multi sub _marshal(@obj) returns Array {
+    multi sub _marshal(Positional:U $, Bool :$skip-null --> Nil ) {
+        Nil;
+    }
+    multi sub _marshal(@obj, Bool :$skip-null --> Array) {
         my @ret;
 
         for @obj -> $item {
-            @ret.push(_marshal($item));
+            @ret.push(_marshal($item, :$skip-null));
         }
         @ret;
     }
-    
-    multi sub _marshal(Mu $obj, Bool :$skip-null) returns Hash {
+
+    multi sub _marshal(Mu:U $, Bool :$skip-null --> Nil ) {
+        Nil;
+    }
+
+    multi sub _marshal(Mu:D $obj, Bool :$skip-null --> Hash ) {
         my %ret;
+        my %local-attrs =  $obj.^attributes(:local).map({ $_.name => $_.package });
         for $obj.^attributes -> $attr {
+            if %local-attrs{$attr.name}:exists && !(%local-attrs{$attr.name} === $attr.package ) {
+                next;
+            }
             if $attr.has_accessor {
                 my $name = do if $attr ~~ JSON::Name::NamedAttribute {
                     $attr.json-name;
@@ -149,17 +175,23 @@ module JSON::Marshal:ver<0.0.6>:auth<github:jonathanstowe> {
         %ret;
     }
 
-    sub serialise-ok(Attribute $attr, $value, Bool $skip-null ) returns Bool {
+    sub serialise-ok(Attribute $attr, $value, Bool $skip-null --> Bool ) {
         my $rc = True;
-        if $skip-null || ( $attr ~~ SkipNull ) {
-            if not $value.defined {
-                $rc = False;
+        if  $attr ~~ JsonSkip {
+            $rc = False;
+        }
+        elsif $skip-null || ( $attr ~~ SkipNull ) {
+            if $attr.type ~~ Associative|Positional {
+                $rc = ?$value.elems;
+            }
+            else {
+                $rc = $value.defined;
             }
         }
         $rc;
     }
 
-    sub marshal(Any $obj, Bool :$skip-null) returns Str is export {
+    sub marshal(Any $obj, Bool :$skip-null --> Str ) is export {
         my $ret = _marshal($obj, :$skip-null);
         to-json($ret);
     }
